/**
 * Copyright (C) 2016 Czech Technical University in Prague
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later
 * version.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more
 * details. You should have received a copy of the GNU General Public License along with this
 * program. If not, see
 * <http://www.gnu.org/licenses/>.
 */
package cz.cvut.kbss.reporting.config;

import cz.cvut.kbss.reporting.security.CsrfHeaderFilter;
import cz.cvut.kbss.reporting.security.SecurityConstants;
import cz.cvut.kbss.reporting.security.oauth2.google.GoogleUserInfoTokenServices;
import cz.cvut.kbss.reporting.service.security.UserDetailsService;
import java.io.IOException;
import java.util.Arrays;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.http.HttpStatus;
import org.springframework.security.config.annotation.method.configuration
    .EnableGlobalMethodSecurity;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.builders.WebSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration
    .WebSecurityConfigurerAdapter;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.oauth2.client.OAuth2ClientContext;
import org.springframework.security.oauth2.client.OAuth2RestOperations;
import org.springframework.security.oauth2.client.OAuth2RestTemplate;
import org.springframework.security.oauth2.client.filter.OAuth2ClientAuthenticationProcessingFilter;
import org.springframework.security.oauth2.client.filter.OAuth2ClientContextFilter;
import org.springframework.security.oauth2.client.resource.OAuth2ProtectedResourceDetails;
import org.springframework.security.oauth2.client.token.grant.code.AuthorizationCodeResourceDetails;
import org.springframework.security.oauth2.common.AuthenticationScheme;
import org.springframework.security.oauth2.config.annotation.web.configuration.EnableOAuth2Client;
import org.springframework.security.web.AuthenticationEntryPoint;
import org.springframework.security.web.access.ExceptionTranslationFilter;
import org.springframework.security.web.access.intercept.FilterSecurityInterceptor;
import org.springframework.security.web.authentication.LoginUrlAuthenticationEntryPoint;
import org.springframework.security.web.authentication.SimpleUrlAuthenticationSuccessHandler;
import org.springframework.security.web.authentication.logout
    .HttpStatusReturningLogoutSuccessHandler;
import org.springframework.security.web.csrf.CsrfFilter;

@Configuration @EnableWebSecurity  // Use basic Spring security configuration
@EnableOAuth2Client @PropertySource("classpath:google-auth2.properties")
@ComponentScan(basePackages = "cz.cvut.kbss.reporting.security")
@EnableGlobalMethodSecurity(prePostEnabled = true, securedEnabled = true)
// Allow methods to be secured using annotation
public class SecurityConfig extends WebSecurityConfigurerAdapter {

    private static final String[] COOKIES_TO_DESTROY =
        {SecurityConstants.SESSION_COOKIE_NAME, SecurityConstants.REMEMBER_ME_COOKIE_NAME,
         SecurityConstants.CSRF_COOKIE_NAME};

    @Autowired private OAuth2ClientContextFilter oauth2ClientContextFilter;

    @Autowired private OAuth2ClientContext oauth2ClientContext;

    @Autowired private UserDetailsService userDetailsService;

    @Autowired private Environment environment;

    private OAuth2ProtectedResourceDetails authorizationCodeResource() {
        AuthorizationCodeResourceDetails details = new AuthorizationCodeResourceDetails();
        details.setId("google-oauth-client");
        details.setClientId(environment.getProperty("oauth2.clientId"));
        details.setClientSecret(environment.getProperty("oauth2.clientSecret"));
        details.setUserAuthorizationUri(environment.getProperty("oauth2.userAuthorizationUri"));
        details.setAccessTokenUri(environment.getProperty("oauth2.accessTokenUri"));
        details.setTokenName(environment.getProperty("oauth2.tokenName"));
        String commaSeparatedScopes = environment.getProperty("oauth2.scope");
        details.setScope(Arrays.asList((String[]) commaSeparatedScopes.split(",")));
        details.setAuthenticationScheme(AuthenticationScheme.query);
        details.setClientAuthenticationScheme(AuthenticationScheme.form);
        return details;
    }

    private OAuth2ClientAuthenticationProcessingFilter oauth2ClientAuthenticationProcessingFilter
        () {
        // Used to obtain access token from authorization server (AS)
        final OAuth2RestOperations restTemplate =
            new OAuth2RestTemplate(authorizationCodeResource(), oauth2ClientContext);
        final OAuth2ClientAuthenticationProcessingFilter filter =
            new OAuth2ClientAuthenticationProcessingFilter(
                environment.getProperty("oauth2.filterCallbackPath"));
        filter.setRestTemplate(restTemplate);
        // Set a service that validates an OAuth2 access token
        // We can use either Google API's UserInfo or TokenInfo
        // For this, we chose to use UserInfo service
        filter.setTokenServices(
            new GoogleUserInfoTokenServices(environment.getProperty("oauth2.userInfoUri"),
                environment.getProperty("oauth2.clientId")));

        filter.setAuthenticationSuccessHandler(new SimpleUrlAuthenticationSuccessHandler() {
            public void onAuthenticationSuccess(HttpServletRequest request,
                                                HttpServletResponse response,
                                                Authentication authentication)
                throws IOException, ServletException {
                userDetailsService.ensureExistsExternalUser( ((cz.cvut.kbss.reporting.security.model.UserDetails) authentication.getPrincipal()).getUser());
                this.setDefaultTargetUrl("/#/dashboard");
                super.onAuthenticationSuccess(request, response, authentication);
            }
        });


        return filter;
    }

    private AuthenticationEntryPoint authenticationEntryPoint() {
        // May need an OAuth2AuthenticationEntryPoint for non-browser clients
        return new LoginUrlAuthenticationEntryPoint(
            environment.getProperty("oauth2.filterCallbackPath"));
    }

    @Override public void configure(WebSecurity web) throws Exception {
        web.ignoring().antMatchers("/", "/resources/**", "/js/**");
    }

    @Override protected void configure(HttpSecurity http) throws Exception {
        http.authorizeRequests().anyRequest().permitAll()
            .and().headers().frameOptions().sameOrigin()
            .and().anonymous().disable().exceptionHandling().authenticationEntryPoint(authenticationEntryPoint())
            .and()
            .addFilterAfter(new CsrfHeaderFilter(), CsrfFilter.class).csrf().disable()
            .addFilterAfter(oauth2ClientContextFilter, ExceptionTranslationFilter.class)
            .addFilterBefore(oauth2ClientAuthenticationProcessingFilter(),
                FilterSecurityInterceptor.class)
            .logout().invalidateHttpSession(true).deleteCookies(COOKIES_TO_DESTROY)
                .logoutUrl("/j_spring_security_logout").logoutSuccessHandler((new HttpStatusReturningLogoutSuccessHandler(HttpStatus.OK)))
            .and().sessionManagement().maximumSessions(1);
    }
}
