package cz.cvut.kbss.reporting.service;

import cz.cvut.kbss.reporting.model.Severity;
import cz.cvut.kbss.reporting.rest.util.RawJson;
import cz.cvut.kbss.reporting.service.data.DataLoader;
import cz.cvut.kbss.reporting.util.Constants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.File;
import java.io.UnsupportedEncodingException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.*;

@Service
public class OptionsService {

    private static final Logger LOG = LoggerFactory.getLogger(OptionsService.class);

    /**
     * URL of the repository from which remote loader loads data.
     * <p>
     * Note: This should be externalized into a config file as part of one of the lab tasks
     */
    private static final String REPOSITORY_URL = "http://dev.inbas.cz/openrdf-sesame/repositories/inbas-model-2016-01-21:22-46-17";

    private final Map<String, String> remoteData = new HashMap<>();
    private final Map<String, String> localData = new HashMap<>();
    private final Set<String> internalData = new HashSet<>();

    private final DataLoader remoteLoader;

    private final DataLoader localLoader;

    private final DataLoader internalLoader;

    @Autowired
    public OptionsService(@Qualifier("remoteDataLoader") DataLoader remoteLoader,
                          @Qualifier("localDataLoader") DataLoader localLoader,
                          @Qualifier("internalDataLoader") DataLoader internalLoader) {
        this.remoteLoader = remoteLoader;
        this.localLoader = localLoader;
        this.internalLoader = internalLoader;
    }

    @PostConstruct
    private void initDataSource() {
        discoverRemoteQueryFiles();
        discoverLocalOptionsFiles();
        internalData.add(Severity.getOptionType());
    }

    private void discoverRemoteQueryFiles() {
        final URL folderUrl = this.getClass().getClassLoader().getResource(Constants.QUERY_FILES_DIRECTORY);
        if (folderUrl == null) {
            LOG.error("Query directory {} not found.", Constants.QUERY_FILES_DIRECTORY);
            return;
        }
        discoverOptionsFiles(folderUrl, remoteData);
    }

    private void discoverOptionsFiles(URL folderUrl, Map<String, String> options) {
        try {
            final File dir = new File(folderUrl.toURI());
            final File[] files = dir.listFiles();
            if (files == null || files.length == 0) {
                LOG.warn("No query files found in directory {}.", dir);
                return;
            }
            for (File f : files) {
                if (f.isDirectory()) {
                    continue;
                }
                String category = f.getName();
                category = category.substring(0, category.lastIndexOf('.'));
                options.put(category, dir.getName() + File.separator + f.getName());
            }
        } catch (URISyntaxException e) {
            LOG.error("Unable to get query files from directory {}.", folderUrl, e);
        }
    }

    private void discoverLocalOptionsFiles() {
        final URL folderUrl = this.getClass().getClassLoader().getResource(Constants.OPTION_FILES_DIRECTORY);
        if (folderUrl == null) {
            LOG.warn("Options directory {} not found.", Constants.OPTION_FILES_DIRECTORY);
            return;
        }
        discoverOptionsFiles(folderUrl, localData);
    }

    /**
     * Returns options of the specified type.
     *
     * @param type Type of the required options
     * @return Options in an appropriate representation (List, RawJson etc.)
     * @throws IllegalArgumentException When unknown option type is provided
     */
    public Object getOptions(String type) {
        if (internalData.contains(type)) {
            return new RawJson(internalLoader.loadData(type, Collections.emptyMap()));
        } else if (localData.containsKey(type)) {
            return loadLocalData(localData.get(type));
        } else if (remoteData.containsKey(type)) {
            return loadRemoteData(remoteData.get(type));
        }
        throw new IllegalArgumentException("Unsupported option type " + type + ".");
    }

    private RawJson loadLocalData(String optionsFile) {
        return new RawJson(localLoader.loadData(optionsFile, Collections.emptyMap()));
    }

    private RawJson loadRemoteData(String queryFile) {
        String query = localLoader.loadData(queryFile, Collections.emptyMap());
        try {
            query = URLEncoder.encode(query, Constants.UTF_8_ENCODING);
            final String data = remoteLoader.loadData(REPOSITORY_URL, Collections.singletonMap("query", query));
            return new RawJson(data);
        } catch (UnsupportedEncodingException e) {
            throw new IllegalStateException("Unable to find encoding " + Constants.UTF_8_ENCODING, e);
        }
    }
}
