package cz.cvut.kbss.reporting.rest;

import com.fasterxml.jackson.core.type.TypeReference;
import cz.cvut.kbss.reporting.dto.OccurrenceReportDto;
import cz.cvut.kbss.reporting.environment.Generator;
import cz.cvut.kbss.reporting.model.AbstractEntity;
import cz.cvut.kbss.reporting.model.OccurrenceReport;
import cz.cvut.kbss.reporting.model.Tag;
import cz.cvut.kbss.reporting.rest.handler.ErrorInfo;
import cz.cvut.kbss.reporting.rest.handler.RestExceptionHandler;
import cz.cvut.kbss.reporting.rest.mapper.DtoMapper;
import cz.cvut.kbss.reporting.rest.mapper.DtoMapperImpl;
import cz.cvut.kbss.reporting.service.repository.OccurrenceReportService;
import cz.cvut.kbss.reporting.service.repository.TagService;
import org.junit.Before;
import org.junit.Test;
import org.mockito.*;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import java.time.LocalDateTime;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

public class ReportControllerTest extends BaseControllerTestRunner {

    @Mock
    private OccurrenceReportService serviceMock;

    @Mock
    private TagService tagServiceMock;

    @Spy
    private final DtoMapper dtoMapper = new DtoMapperImpl();

    @InjectMocks
    private ReportController controller;

    @Before
    public void setUp() {
        MockitoAnnotations.initMocks(this);
        super.setUp();
        // This sets up only the single controller, no spring context is actually started
        // setControllerAdvice makes sure that our ExceptionHandler works
        this.mockMvc = MockMvcBuilders.standaloneSetup(controller).setControllerAdvice(new RestExceptionHandler())
                                      .build();
    }

    @Test
    public void findAllReturnsDtosWithRestrictedInformation() throws Exception {
        final List<OccurrenceReport> reports = generateReports();
        when(serviceMock.findAll()).thenReturn(reports);
        final MvcResult mvcResult = mockMvc.perform(get("/reports")).andExpect(status().isOk()).andReturn();
        final List<OccurrenceReportDto> result = readValue(mvcResult, new TypeReference<List<OccurrenceReportDto>>() {
        });

        assertEquals(reports.size(), result.size());
        for (int i = 0; i < reports.size(); i++) {
            final OccurrenceReport expected = reports.get(i);
            final OccurrenceReportDto actual = result.get(i);
            assertEquals(expected.getId(), actual.getId());
            assertEquals(expected.getCreated(), actual.getCreated());
            assertEquals(expected.getOccurrence().getId(), actual.getOccurrence().getId());
            assertEquals(expected.getOccurrence().getName(), actual.getOccurrence().getName());
            assertEquals(expected.getOccurrence().getStartTime(), actual.getOccurrence().getStartTime());
        }
    }

    private List<OccurrenceReport> generateReports() {
        return IntStream.range(0, Generator.randomInt(5, 10)).mapToObj(i -> {
            final OccurrenceReport r = Generator.generateReport();
            r.setId(i);
            r.getOccurrence().setId(i);
            r.setCreated(LocalDateTime.now());
            return r;
        }).collect(Collectors.toList());
    }

    @Test
    public void findReturnsNotFoundForUnknownReportId() throws Exception {
        when(serviceMock.find(any())).thenReturn(null);
        final MvcResult mvcResult = mockMvc.perform(get("/reports/" + 117)).andExpect(status().isNotFound())
                                           .andReturn();
        final ErrorInfo errorInfo = readValue(mvcResult, ErrorInfo.class);
        assertNotNull(errorInfo);
        assertThat(errorInfo.getMessage(), containsString("with id " + 117 + " not found."));
    }

    @Test
    public void createReportPassesDataToService() throws Exception {
        final OccurrenceReport newReport = Generator.generateReport();
        mockMvc.perform(post("/reports").content(toJson(newReport)).contentType(MediaType.APPLICATION_JSON_VALUE))
               .andExpect(status().isCreated());
        final ArgumentCaptor<OccurrenceReport> captor = ArgumentCaptor.forClass(OccurrenceReport.class);
        verify(serviceMock).persist(captor.capture());
        final OccurrenceReport report = captor.getValue();
        assertEquals(newReport.getNarrative(), report.getNarrative());
        assertEquals(newReport.getOccurrence().getType(), report.getOccurrence().getType());
    }

    @Test
    public void createReportReturnsLocationHeaderWithReportUri() throws Exception {
        final OccurrenceReport newReport = Generator.generateReport();
        newReport.setId(117);
        MvcResult mvcResult = mockMvc
                .perform(post("/reports").content(toJson(newReport)).contentType(MediaType.APPLICATION_JSON_VALUE))
                .andExpect(status().isCreated()).andReturn();
        final String location = mvcResult.getResponse().getHeader(HttpHeaders.LOCATION);
        assertNotNull(location);
        assertThat(location, containsString("reports/" + newReport.getId()));
    }

    @Test
    public void deleteReportRemovesReport() throws Exception {
        final OccurrenceReport report = Generator.generateReport();
        report.setId(117);
        when(serviceMock.find(report.getId())).thenReturn(report);
        mockMvc.perform(delete("/reports/" + report.getId())).andExpect(status().isNoContent());
        verify(serviceMock).find(report.getId());
        verify(serviceMock).remove(report);
    }

    @Test
    public void deleteReportThrowsNotFoundForUnknownReportId() throws Exception {
        final Integer unknownId = 118;
        final MvcResult mvcResult = mockMvc.perform(delete("/reports/" + unknownId)).andExpect(status().isNotFound())
                                           .andReturn();
        final ErrorInfo error = readValue(mvcResult, ErrorInfo.class);
        assertThat(error.getMessage(), containsString(unknownId + " not found"));
        verify(serviceMock, never()).remove(any());
    }

    @Test
    public void updateUpdatesReport() throws Exception {
        final OccurrenceReport report = Generator.generateReport();
        report.setId(117);
        when(serviceMock.exists(report.getId())).thenReturn(true);
        mockMvc.perform(
                put("/reports/" + report.getId()).content(toJson(report)).contentType(MediaType.APPLICATION_JSON_VALUE))
               .andExpect(status().isNoContent());
        final ArgumentCaptor<OccurrenceReport> captor = ArgumentCaptor.forClass(OccurrenceReport.class);
        verify(serviceMock).update(captor.capture());
        final OccurrenceReport argument = captor.getValue();
        assertEquals(report.getId(), argument.getId());
        assertEquals(report.getNarrative(), argument.getNarrative());
        assertEquals(report.getOccurrence().getType(), argument.getOccurrence().getType());
    }

    @Test
    public void updateThrowsNotFoundForUnknownReportId() throws Exception {
        final OccurrenceReport report = Generator.generateReport();
        report.setId(117);
        when(serviceMock.exists(report.getId())).thenReturn(false);
        final MvcResult mvcResult = mockMvc.perform(
                put("/reports/" + report.getId()).content(toJson(report)).contentType(MediaType.APPLICATION_JSON_VALUE))
                                           .andExpect(status().isNotFound()).andReturn();
        final ErrorInfo error = readValue(mvcResult, ErrorInfo.class);
        assertThat(error.getMessage(), containsString(report.getId() + " not found"));
        verify(serviceMock, never()).update(any());
    }

    @Test
    public void updatedThrowsConflictWhenReportIdDoesNotMatchPassedData() throws Exception {
        final OccurrenceReport report = Generator.generateReport();
        report.setId(117);
        final Integer otherId = 118;
        mockMvc.perform(
                put("/reports/" + otherId).content(toJson(report)).contentType(MediaType.APPLICATION_JSON_VALUE))
               .andExpect(status().isConflict());
        verify(serviceMock, never()).exists(any());
        verify(serviceMock, never()).update(any());
    }

    @Test
    public void findAllWithTagQueryParameterReturnsReportsWithMatchingTag() throws Exception {
        final List<OccurrenceReport> reports = generateReports();
        final List<OccurrenceReport> reportsWithTag = reports.stream().filter(r -> Generator.randomBoolean()).collect(
                Collectors.toList());
        when(serviceMock.findAll()).thenReturn(reports);
        final String tagName = "taggard";
        final Tag tag = Generator.generateTag(tagName);
        when(tagServiceMock.findByName(tagName)).thenReturn(tag);
        when(serviceMock.findAll(tag)).thenReturn(reportsWithTag);

        final MvcResult mvcResult = mockMvc.perform(get("/reports").param("tag", tagName)).andExpect(status().isOk())
                                           .andReturn();
        final List<OccurrenceReportDto> result = readValue(mvcResult, new TypeReference<List<OccurrenceReportDto>>() {
        });
        final Set<Integer> reportIds = reportsWithTag.stream().map(AbstractEntity::getId).collect(Collectors.toSet());
        result.forEach(r -> assertTrue(reportIds.contains(r.getId())));
        verify(tagServiceMock).findByName(tagName);
        verify(serviceMock).findAll(tag);
    }

    @Test
    public void findAllWithTagThrowsNotFoundForUnknownTag() throws Exception {
        final String unknownTag = "unknownTag";
        when(tagServiceMock.findByName(unknownTag)).thenReturn(null);
        final MvcResult mvcResult = mockMvc.perform(get("/reports").param("tag", unknownTag))
                                           .andExpect(status().isNotFound())
                                           .andReturn();
        final ErrorInfo errorInfo = readValue(mvcResult, ErrorInfo.class);
        assertThat(errorInfo.getMessage(), containsString("Tag " + unknownTag + " not found."));
        verify(serviceMock, never()).findAll(any());
        verify(serviceMock, never()).findAll();
    }
}